import sys

sys.path.append("src")

from mamba4rec import Dataset
import pytest
from collections.abc import Iterable


class DatasetML1M(Dataset):
    def __init__(self, path_to_dataset: str, leave_k_out: int = 5):
        self._path_to_dataset = path_to_dataset
        self._leave_k_out = leave_k_out
        super().__init__(self._load_interactions(path_to_dataset), leave_k_out)

    def _load_interactions(self, path_to_dataset) -> Iterable[tuple[int, int]]:
        path_to_interactions = path_to_dataset + "/" + "ratings.dat"
        with open(path_to_interactions, "r") as file:
            for line in file:
                yield tuple(map(int, line.strip("\n").split("::")[:2]))


@pytest.fixture
def dataset_ml_1m():
    return DatasetML1M("dataset/ml-1m", leave_k_out=5)


def test_item2id(dataset_ml_1m):
    assert len(dataset_ml_1m.item2id) > 1000


def test_user2id(dataset_ml_1m):
    assert len(dataset_ml_1m.id2user) > 1000


def test_id2item(dataset_ml_1m):
    assert len(dataset_ml_1m.id2item) > 1000


def test_val_item_ids(dataset_ml_1m):
    assert len(dataset_ml_1m.val_item_ids) == len(dataset_ml_1m.id2user)


def test_train_item_ids(dataset_ml_1m):
    assert len(dataset_ml_1m.train_item_ids) == len(dataset_ml_1m.id2user)


def test_train_item_ids(dataset_ml_1m):
    assert len(dataset_ml_1m.train_item_ids[512]) >= dataset_ml_1m.leave_k_out


def test_val_item_ids(dataset_ml_1m):
    assert len(dataset_ml_1m.val_item_ids[512]) == dataset_ml_1m.leave_k_out



import argparse
from train import TrainModel, Datasets, Vocab
from s3_tools import s3_tools
from dill import dumps

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-bn",
        "--bucket_name",
        type=str,
        required=True,
        help="Bucket S3 dataset",
    )
    parser.add_argument(
        "-dkn",
        "--data_key_name",
        type=str,
        required=True,
        help="Path to S3 object",
    )
    parser.add_argument(
        "-mfn",
        "--model_folder_name",
        type=str,
        required=True,
        help="Path to S3 model folder",
    )

    args = parser.parse_args()
    print(vars(args), flush=True)

    s3 = s3_tools()
    data_dict = s3.get_dill_object(
        bucket_name=args.bucket_name, key_name=args.data_key_name
    )
    vocab = Vocab(data_dict.get("search_texts", set()))
    TrainModel(
        vocab,
        Datasets(
            data_dict.get("train_interactions", []),
            data_dict.get("test_interactions", []),
        )
    )

    s3.safe_upload_folder(
        folder_name="./saved/*",
        bucket_name=args.bucket_name,
        object_name=args.model_folder_name.strip("/") + "/",
    )

    s3.s3_client.put_object(
        Bucket=args.bucket_name,
        Key=args.model_folder_name.strip("/") + "/" + "vocab.obj",
        Body=dumps(
            {
                key: value
                for key, value in data_dict.items()
                if "interactions" not in key
            }
        ),
    )



import torch
from sklearn.model_selection import train_test_split
from torch.nn.utils.rnn import pad_sequence
from transformers import MambaConfig, MambaForCausalLM, Trainer, TrainingArguments


class ListDataset(torch.utils.data.Dataset):
    def __init__(self, data: list):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

    def __getitems__(self, idx_list):
        return [self.data[_] for _ in idx_list]


class DataCollatorForCLMRec:
    def __init__(self, pad_id):
        self.pad_id = pad_id

    def mask_ids_batch(self, batch_of_ids):

        padded_ids = pad_sequence(
            list(map(lambda ids: torch.LongTensor(ids[::-1]), batch_of_ids)),
            batch_first=True,
            padding_value=self.pad_id,
        ).flip(dims=[1])
        return {
            "input_ids": padded_ids,
            "attention_mask": padded_ids != self.pad_id,
            "labels": padded_ids,
        }

    def __call__(self, batch_of_ids):
        return self.mask_ids_batch(batch_of_ids)


class Vocab:
    def __init__(self, items: set):

        self._item2id = {item: idx for idx, item in enumerate(items)}
        self._item2id[self.pad_str] = len(self._item2id)
        self._item2id[self.unk_str] = len(self._item2id)

    @property
    def vocab_size(self) -> int:
        return len(self._item2id)

    @property
    def id2item(self) -> dict:
        return list(self._item2id.keys())

    @property
    def item2id(self) -> dict:
        return self._item2id

    @property
    def pad_str(self) -> str:
        return "[PAD]"

    @property
    def pad_id(self) -> int:
        return self._item2id.get(self.pad_str, -1)

    @property
    def unk_str(self) -> str:
        return "[UNK]"

    @property
    def unk_id(self) -> int:
        return self._item2id.get(self.unk_str, -1)


class Datasets:
    def __init__(self, train_interactions: list, test_interactions: list):

        X_train, X_test, self._val_train, self._val_test = train_test_split(
            train_interactions,
            test_interactions,
            test_size=0.05,
            random_state=42,
        )

        self._train_dataset = ListDataset(X_train)
        self._eval_dataset = ListDataset(X_test)

    @property
    def train_dataset(self) -> ListDataset:
        return self._train_dataset

    @property
    def eval_dataset(self) -> ListDataset:
        return self._eval_dataset


class TrainModel:
    def __init__(self, vocab: Vocab, dl: Datasets):

        assert len(dl.train_dataset) > len(dl.eval_dataset)

        device = "cuda" if torch.cuda.is_available() else "cpu"

        # -hs 128 -ss 16 -is 64 -hl 8
        config = MambaConfig(
            hidden_size=32,
            num_hidden_layers=8,
            vocab_size=vocab.vocab_size,
            state_size=8,
            intermediate_size=32,
            use_mambapy=True,
            use_cache=False,
            pad_token_id=vocab.pad_id,
            bos_token_id=vocab.pad_id,  ## CLS
            eos_token_id=vocab.pad_id,  ## SEP
            expand=1,
        )
        model = MambaForCausalLM(config).to(device)

        assert model.num_parameters() > 1000

        training_args = TrainingArguments(
            output_dir="./results",
            eval_strategy="steps",
            prediction_loss_only=True,
            save_strategy="best",
            load_best_model_at_end=True,
            metric_for_best_model="eval_loss",
            learning_rate=2e-5,
            per_device_train_batch_size=1024,
            per_device_eval_batch_size=1024,
            num_train_epochs=10,
            weight_decay=0.01,
            use_cpu=False,
            data_seed=42,
            seed=42,
            disable_tqdm=False,
            full_determinism=True,
            save_total_limit=5,
            save_safetensors=False,
        )

        trainer = Trainer(
            model=model,
            args=training_args,
            data_collator=DataCollatorForCLMRec(vocab.pad_id),
            train_dataset=dl.train_dataset,
            eval_dataset=dl.eval_dataset,
        )

        trainer.train()
        trainer.save_model("./saved")



import argparse

import torch
from mamba4rec.s3_tools import s3_tools
from sklearn.metrics import ndcg_score
from torch.utils.data import DataLoader
from tqdm import tqdm
from mamba4rec.train import DataCollatorForCLMRec, ListDataset
from transformers import MambaForCausalLM
from transformers.generation.configuration_utils import GenerationConfig

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-bn",
        "--bucket_name",
        type=str,
        required=True,
        help="Bucket S3 dataset",
    )
    parser.add_argument(
        "-dkn",
        "--data_key_name",
        type=str,
        required=True,
        help="Path to S3 object",
    )

    parser.add_argument(
        "-mfn",
        "--model_folder_name",
        type=str,
        required=True,
        help="Path to S3 object model",
    )
    parser.add_argument(
        "-bs",
        "--batch_size",
        type=int,
        required=True,
        help="batch size for inference",
        default=12,
    )

    args = parser.parse_args()
    print(vars(args), flush=True)

    s3 = s3_tools()
    data_dict = s3.get_dill_object(
        bucket_name=args.bucket_name, key_name=args.data_key_name
    )
    print(data_dict.keys())

    at_k = len(data_dict.get("test_interactions", [])[0])

    ##check if exist local folder or get S3
    s3.download_folder(
        bucket_name=args.bucket_name,
        object_name=args.model_folder_name,
        folder_name="./saved/",
    )
    model = MambaForCausalLM.from_pretrained("./saved/")
    pad_id = model.config.pad_token_id

    device = "cuda" if torch.cuda.is_available() else "cpu"
    model.to(device)
    model.eval()

    gconf = GenerationConfig(
        max_new_tokens=at_k,
        num_beams=1,
        do_sample=True,
        pad_token_id=pad_id,
        no_repeat_ngram_size=1,
    )

    train_inference = []
    train_dataloader = DataLoader(
        ListDataset(data_dict.get("train_interactions", [])),
        batch_size=args.batch_size,
        collate_fn=DataCollatorForCLMRec(pad_id),
        shuffle=False,
    )
    with torch.no_grad():
        for batch in tqdm(train_dataloader):

            train_inference += (
                model.generate(
                    batch["input_ids"].to(device),
                    attention_mask=batch["attention_mask"].to(device),
                    generation_config=gconf,
                )
                .detach()
                .cpu()[:, -at_k:]
                .tolist()
            )

    print(
        f'nDCG@{at_k} = {ndcg_score(data_dict.get("test_interactions", []), train_inference):.4f}'
    )




from .test_ml_1m import  DatasetML1M 
import pytest

import torch
from torch.nn.utils.rnn import pad_sequence
from transformers import MambaConfig, MambaForCausalLM
from transformers import TrainingArguments, Trainer
from sklearn.model_selection import train_test_split
import os
import random

def seed_everything(seed):
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cuda.matmul.allow_tf32 = True

class ListDataset(torch.utils.data.Dataset):
    def __init__(self, data: list):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

    def __getitems__(self, idx_list):
        return [self.data[_] for _ in idx_list]

class DataCollatorForCLMRec:
    def __init__(self, pad_id):
        self.pad_id = pad_id

    def mask_ids_batch(self, batch_of_ids):
        
        padded_ids = pad_sequence(
            list(map(lambda ids: torch.LongTensor(ids[::-1]), batch_of_ids)),
            batch_first=True,
            padding_value=self.pad_id,
        ).flip(dims=[1])
        #labels = padded_ids[:, -1]#.unsqueeze(dim=1)
        #padded_ids = padded_ids[:,:-1]
        #print(f"{padded_ids.shape=} {labels.shape=}")
        return {
            "input_ids": padded_ids,
            "attention_mask": padded_ids != self.pad_id,
            "labels": padded_ids,
        }

    def __call__(self, batch_of_ids):
        return self.mask_ids_batch(batch_of_ids)

class CreateDataloaders:
    def __init__(self, dataset : DatasetML1M ):
        self._dataset = dataset
        self._pad_str = "[PAD]"
        self._pad_id = len(self._dataset.item2id)
        self._dataset.item2id[self._pad_str] = self._pad_id
        self._unk_id = len(self._dataset.item2id)
        self._unk_str = "[UNK]"
        self._dataset.item2id[self._unk_str] = self._unk_id
        self._vocab_size = len(self._dataset.item2id)


        X_train, X_test, self._val_train, self._val_test = train_test_split(self._dataset.train_item_ids, self._dataset.val_item_ids,  test_size=0.05, random_state=42)

        self._train_dataset = ListDataset(X_train)
        self._test_dataset = ListDataset(X_test)

    def save_model(self):
        #self._fast_tokenizer = PreTrainedTokenizerFast(tokenizer_object = Tokenizer(WordLevel(self._dataset.item2id, self._unk_str)),   pad_token=self._pad_str, unk_token=self._unk_str,)
        #fast_tokenizer.save_pretrained(args.hf_path)
        #model.save_pretrained(args.hf_path + "/last")
        #fast_tokenizer.save_pretrained(args.hf_path + "/last")      
        pass

@pytest.fixture
def dataset_ml_1m():
    return DatasetML1M("dataset/ml-1m", leave_k_out=2)

def test_model_train(dataset_ml_1m):

    dl = CreateDataloaders(dataset_ml_1m)
    assert len(dl._train_dataset) > len (dl._test_dataset)

    device = "cuda" if torch.cuda.is_available() else "cpu"

    config = MambaConfig(
        hidden_size=8,
        num_hidden_layers=2,
        vocab_size=dl._vocab_size,
        state_size=4,
        intermediate_size=4,
        use_mambapy=True,
        use_cache=False,
        pad_token_id=dl._pad_id,
        bos_token_id=dl._pad_id,  ## CLS
        eos_token_id=dl._pad_id,  ## SEP
        expand=1,
    )
    #print(config, flush=True)
    model = MambaForCausalLM(config).to(device)
    
    assert model.num_parameters() > 1000

    #print(model.__class__, model.num_parameters(), "parameters!", flush=True)
    #print(model, flush=True)

    training_args = TrainingArguments(
        output_dir="./results",
        eval_strategy="epoch",
        learning_rate=2e-5,
        per_device_train_batch_size=4,
        per_device_eval_batch_size=4,
        num_train_epochs=1,
        weight_decay=0.01,
        use_cpu = True,
        data_seed = 42,
        seed = 42,
        disable_tqdm = True
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        data_collator=DataCollatorForCLMRec(dl._pad_id),
        train_dataset=dl._train_dataset,
        eval_dataset=dl._test_dataset,
    )
        
    assert trainer.train()
    assert trainer.evaluate()


